@@ -50,8 +50,9 @@ class ChessState:
 
 def init():
     bs = np.zeros(64, dtype=np.int32)
-    for i in range(8):
-        bs[1 + 8 * i] = 1
+    np.put(bs, np.arange(1, 65, 8), 1)
+    # for i in range(8):
+    #    bs[1 + 8 * i] = 1
     bs[8] = 2
     bs[48] = 2
     bs[16] = 3
@@ -60,8 +61,9 @@ def init():
     bs[56] = 4
     bs[24] = 5
     bs[32] = 6
-    for i in range(8):
-        bs[6 + 8 * i] = 7
+    np.put(bs, np.arange(6, 70, 8), 7)
+    # for i in range(8):
+    #    bs[6 + 8 * i] = 7
     bs[15] = 8
     bs[55] = 8
     bs[23] = 9
@@ -315,6 +317,50 @@ def dif_to_direction(from_: int, to: int) -> Tuple[int, int]:
     return -1, 0
 
 
+def _is_same_line(from_: int, to: int, direction: int):
+    if direction <= 1:
+        return _is_same_column(from_, to)
+    elif direction <= 3:
+        return _is_same_row(from_, to)
+    elif direction <= 5:
+        return _is_same_rising(from_, to)
+    elif direction <= 7:
+        return _is_same_declining(from_, to)
+    else:
+        return False
+
+
+def _dis_direction_array(from_: int, direction: int):
+    array = np.zeros(64, dtype=np.int32)
+    dif = _direction_to_dif(direction)
+    to = from_ + dif
+    # for i in range(7):
+    #    to_ = from_ + dif * (1 + i)
+    #    if _is_in_board(to_) and _is_same_line(from_, to_, direction):
+    #        array[to_] = i + 1
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 1
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 2
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 3
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 4
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 5
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 6
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 7
+    return array
+
+
 def _piece_type(state: ChessState, position: int) -> int:
     return int(state.board[:, position].argmax())
 
@@ -425,10 +471,7 @@ def _is_same_declining(_from: int, to: int) -> bool:
 
 
 def _board_status(state: ChessState) -> np.ndarray:
-    bs = np.zeros(64, dtype=np.int32)
-    for i in range(64):
-        bs[i] = _piece_type(state, i)
-    return bs
+    return state.board.argmax(axis=0)
 
 
 def _owner(piece: int) -> int:
@@ -513,14 +556,10 @@ def _white_pawn_moves(bs: np.ndarray, from_: int, pin: int) -> np.ndarray:
     if _is_in_board(from_ - 7) and (pin == 0 or pin == 4):
         if _owner(bs[from_ - 7]) == 1:
             to[from_ - 7] = 1
-        elif _owner(bs[from_ - 7]) == 0:
-            to[from_ - 7] = 2
     # 右斜め前
     if _is_in_board(from_ + 9) and (pin == 0 or pin == 3):
         if _owner(bs[from_ + 9]) == 1:
             to[from_ + 9] = 1
-        elif _owner(bs[from_ + 9]) == 0:
-            to[from_ + 9] = 2
     return to
 
 
@@ -601,28 +640,57 @@ def _knight_moves(
     su, sd, sl, sr = _is_second_line(from_)
     # 上方向
     if not su:
-        if not l_:
-            to[from_ - 6] = _piece_turn_to_color(bs[from_ - 6], turn)
-        if not r:
-            to[from_ + 10] = _piece_turn_to_color(bs[from_ + 10], turn)
+        if not l_ and _owner(bs[from_ - 6]) != turn:
+            to[from_ - 6] = 1
+        if not r and _owner(bs[from_ + 10]) != turn:
+            to[from_ + 10] = 1
     # 左方向
     if not sl:
-        if not u:
-            to[from_ - 15] = _piece_turn_to_color(bs[from_ - 15], turn)
-        if not d:
-            to[from_ - 17] = _piece_turn_to_color(bs[from_ - 17], turn)
+        if not u and _owner(bs[from_ - 15]) != turn:
+            to[from_ - 15] = 1
+        if not d and _owner(bs[from_ - 17]) != turn:
+            to[from_ - 17] = 1
     # 下方向
     if not sd:
-        if not l_:
-            to[from_ - 10] = _piece_turn_to_color(bs[from_ - 10], turn)
-        if not r:
-            to[from_ + 6] = _piece_turn_to_color(bs[from_ + 6], turn)
+        if not l_ and _owner(bs[from_ - 10]) != turn:
+            to[from_ - 10] = 1
+        if not r and _owner(bs[from_ + 6]) != turn:
+            to[from_ + 6] = 1
     # 右方向
     if not sr:
+        if not u and _owner(bs[from_ + 17]) != turn:
+            to[from_ + 17] = 1
+        if not d and _owner(bs[from_ + 15]) != turn:
+            to[from_ + 15] = 1
+    return to
+
+
+def _knight_effect(from_: int):
+    to = np.zeros(64, dtype=np.int32)
+    # pinされている場合は動けない
+    u, d, l_, r = _is_side(from_)
+    su, sd, sl, sr = _is_second_line(from_)
+    # 上方向
+    if not su:
+        if not l_:
+            to[from_ - 6] = 1
+        if not r:
+            to[from_ + 10] = 1
+    if not sl:
         if not u:
-            to[from_ + 17] = _piece_turn_to_color(bs[from_ + 17], turn)
+            to[from_ - 15] = 1
         if not d:
-            to[from_ + 15] = _piece_turn_to_color(bs[from_ + 15], turn)
+            to[from_ - 17] = 1
+    if not sd:
+        if not l_:
+            to[from_ - 10] = 1
+        if not r:
+            to[from_ + 6] = 1
+    if not sr:
+        if not u:
+            to[from_ + 17] = 1
+        if not d:
+            to[from_ + 15] = 1
     return to
 
 
@@ -635,39 +703,106 @@ def _bishop_moves(
     ul_flag = pin == 0 or pin == 4
     dr_flag = pin == 0 or pin == 4
     dl_flag = pin == 0 or pin == 3
-    for i in range(8):
-        ur = from_ + 9 * (1 + i)
-        ul = from_ - 7 * (1 + i)
-        dr = from_ + 7 * (1 + i)
-        dl = from_ - 9 * (1 + i)
-        if ur_flag and _is_in_board(ur) and _is_same_rising(from_, ur):
-            if _owner(bs[ur]) == turn:
-                to[ur] = 2
+    ur_array = _dis_direction_array(from_, 5)
+    ul_array = _dis_direction_array(from_, 6)
+    dr_array = _dis_direction_array(from_, 7)
+    dl_array = _dis_direction_array(from_, 4)
+    bs_one = np.where(bs == 0, 0, 1)
+    if ur_flag:
+        if np.all(bs_one * ur_array == 0):
+            if np.all(ur_array == 0):
+                max_dis = 0
             else:
-                to[ur] = 1
-        if ul_flag and _is_in_board(ul) and _is_same_declining(from_, ul):
-            if _owner(bs[ul]) == turn:
-                to[ul] = 2
+                max_dis = np.max(ur_array)
+        else:
+            max_dis = np.min(ur_array[np.nonzero(ur_array * bs_one)])
+            if _owner(bs[from_ + 9 * max_dis]) == turn:
+                max_dis -= 1
+        ur_point = from_ + 9 * max_dis
+        np.put(to, np.arange(ur_point, from_, -9), 1)
+    if ul_flag:
+        if np.all(bs_one * ul_array == 0):
+            if np.all(ul_array == 0):
+                max_dis = 0
             else:
-                to[ul] = 1
-        if dr_flag and _is_in_board(dr) and _is_same_declining(from_, dr):
-            if _owner(bs[dr]) == turn:
-                to[dr] = 2
+                max_dis = np.max(ul_array)
+        else:
+            max_dis = np.min(ul_array[np.nonzero(ul_array * bs_one)])
+            if _owner(bs[from_ - 7 * max_dis]) == turn:
+                max_dis -= 1
+        ul_point = from_ - 7 * max_dis
+        np.put(to, np.arange(ul_point, from_, 7), 1)
+    if dr_flag:
+        if np.all(bs_one * dr_array == 0):
+            if np.all(dr_array == 0):
+                max_dis = 0
             else:
-                to[dr] = 1
-        if dl_flag and _is_in_board(dl) and _is_same_rising(from_, dl):
-            if _owner(bs[dl]) == turn:
-                to[dl] = 2
+                max_dis = np.max(dr_array)
+        else:
+            max_dis = np.min(dr_array[np.nonzero(dr_array * bs_one)])
+            if _owner(bs[from_ + 7 * max_dis]) == turn:
+                max_dis -= 1
+        dr_point = from_ + 7 * max_dis
+        np.put(to, np.arange(dr_point, from_, -7), 1)
+    if dl_flag:
+        if np.all(bs_one * dl_array == 0):
+            if np.all(dl_array == 0):
+                max_dis = 0
             else:
-                to[dl] = 1
-        if not _is_in_board(ur) or bs[ur] != 0:
-            ur_flag = False
-        if not _is_in_board(ul) or bs[ul] != 0:
-            ul_flag = False
-        if not _is_in_board(dr) or bs[dr] != 0:
-            dr_flag = False
-        if not _is_in_board(dl) or bs[dl] != 0:
-            dl_flag = False
+                max_dis = np.max(dl_array)
+        else:
+            max_dis = np.min(dl_array[np.nonzero(dl_array * bs_one)])
+            if _owner(bs[from_ - 9 * max_dis]) == turn:
+                max_dis -= 1
+        dl_point = from_ - 9 * max_dis
+        np.put(to, np.arange(dl_point, from_, 9), 1)
+    return to
+
+
+def _bishop_effect(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = np.zeros(64, dtype=np.int32)
+    ur_array = _dis_direction_array(from_, 5)
+    ul_array = _dis_direction_array(from_, 6)
+    dr_array = _dis_direction_array(from_, 7)
+    dl_array = _dis_direction_array(from_, 4)
+    bs_one = np.where(bs == 0, 0, 1)
+    if np.all(bs_one * ur_array == 0):
+        # ur_arrayに被る駒が存在しない
+        if np.all(ur_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(ur_array)
+    else:
+        max_dis = np.min(ur_array[np.nonzero(ur_array * bs_one)])
+    ur_point = from_ + 9 * max_dis
+    np.put(to, np.arange(ur_point, from_, -9), 1)
+    if np.all(bs_one * ul_array == 0):
+        if np.all(ul_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(ul_array)
+    else:
+        max_dis = np.min(ul_array[np.nonzero(ul_array * bs_one)])
+    ul_point = from_ - 7 * max_dis
+    np.put(to, np.arange(ul_point, from_, 7), 1)
+    if np.all(bs_one * dr_array == 0):
+        if np.all(dr_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(dr_array)
+    else:
+        max_dis = np.min(dr_array[np.nonzero(dr_array * bs_one)])
+    dr_point = from_ + 7 * max_dis
+    np.put(to, np.arange(dr_point, from_, -7), 1)
+    if np.all(bs_one * dl_array == 0):
+        if np.all(dl_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(dl_array)
+    else:
+        max_dis = np.min(dl_array[np.nonzero(dl_array * bs_one)])
+    dl_point = from_ - 9 * max_dis
+    np.put(to, np.arange(dl_point, from_, 9), 1)
     return to
 
 
@@ -677,39 +812,105 @@ def _rook_moves(bs: np.ndarray, from_: int, turn: int, pin: int) -> np.ndarray:
     d_flag = pin == 0 or pin == 1
     r_flag = pin == 0 or pin == 2
     l_flag = pin == 0 or pin == 2
-    for i in range(8):
-        u = from_ + 1 * (1 + i)
-        d = from_ - 1 * (1 + i)
-        l_ = from_ - 8 * (1 + i)
-        r = from_ + 8 * (1 + i)
-        if u_flag and _is_in_board(u) and _is_same_column(from_, u):
-            if _owner(bs[u]) == turn:
-                to[u] = 2
+    u_array = _dis_direction_array(from_, 1)
+    d_array = _dis_direction_array(from_, 0)
+    r_array = _dis_direction_array(from_, 3)
+    l_array = _dis_direction_array(from_, 2)
+    bs_one = np.where(bs == 0, 0, 1)
+    if u_flag:
+        if np.all(bs_one * u_array == 0):
+            if np.all(u_array == 0):
+                max_dis = 0
             else:
-                to[u] = 1
-        if d_flag and _is_in_board(d) and _is_same_column(from_, d):
-            if _owner(bs[d]) == turn:
-                to[d] = 2
+                max_dis = np.max(u_array)
+        else:
+            max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
+            if _owner(bs[from_ + max_dis]) == turn:
+                max_dis -= 1
+        u_point = from_ + max_dis
+        np.put(to, np.arange(u_point, from_, -1), 1)
+    if d_flag:
+        if np.all(bs_one * d_array == 0):
+            if np.all(d_array == 0):
+                max_dis = 0
             else:
-                to[d] = 1
-        if l_flag and _is_in_board(l_) and _is_same_row(from_, l_):
-            if _owner(bs[l_]) == turn:
-                to[l_] = 2
+                max_dis = np.max(d_array)
+        else:
+            max_dis = np.min(d_array[np.nonzero(d_array * bs_one)])
+            if _owner(bs[from_ - max_dis]) == turn:
+                max_dis -= 1
+        d_point = from_ - max_dis
+        np.put(to, np.arange(d_point, from_, 1), 1)
+    if r_flag:
+        if np.all(bs_one * r_array == 0):
+            if np.all(r_array == 0):
+                max_dis = 0
             else:
-                to[l_] = 1
-        if r_flag and _is_in_board(r) and _is_same_row(from_, r):
-            if _owner(bs[r]) == turn:
-                to[r] = 2
+                max_dis = np.max(r_array)
+        else:
+            max_dis = np.min(r_array[np.nonzero(r_array * bs_one)])
+            if _owner(bs[from_ + 8 * max_dis]) == turn:
+                max_dis -= 1
+        r_point = from_ + 8 * max_dis
+        np.put(to, np.arange(r_point, from_, -8), 1)
+    if l_flag:
+        if np.all(bs_one * l_array == 0):
+            if np.all(l_array == 0):
+                max_dis = 0
             else:
-                to[r] = 1
-        if not _is_in_board(u) or bs[u] != 0:
-            u_flag = False
-        if not _is_in_board(d) or bs[d] != 0:
-            d_flag = False
-        if not _is_in_board(l_) or bs[l_] != 0:
-            l_flag = False
-        if not _is_in_board(r) or bs[r] != 0:
-            r_flag = False
+                max_dis = np.max(l_array)
+        else:
+            max_dis = np.min(l_array[np.nonzero(l_array * bs_one)])
+            if _owner(bs[from_ - 8 * max_dis]) == turn:
+                max_dis -= 1
+        l_point = from_ - 8 * max_dis
+        np.put(to, np.arange(l_point, from_, 8), 1)
+    return to
+
+
+def _rook_effect(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = np.zeros(64, dtype=np.int32)
+    u_array = _dis_direction_array(from_, 1)
+    d_array = _dis_direction_array(from_, 0)
+    r_array = _dis_direction_array(from_, 3)
+    l_array = _dis_direction_array(from_, 2)
+    bs_one = np.where(bs == 0, 0, 1)
+    if np.all(bs_one * u_array == 0):
+        if np.all(u_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(u_array)
+    else:
+        max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
+    u_point = from_ + max_dis
+    np.put(to, np.arange(u_point, from_, -1), 1)
+    if np.all(bs_one * d_array == 0):
+        if np.all(d_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(d_array)
+    else:
+        max_dis = np.min(d_array[np.nonzero(d_array * bs_one)])
+    d_point = from_ - max_dis
+    np.put(to, np.arange(d_point, from_, 1), 1)
+    if np.all(bs_one * r_array == 0):
+        if np.all(r_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(r_array)
+    else:
+        max_dis = np.min(r_array[np.nonzero(r_array * bs_one)])
+    r_point = from_ + 8 * max_dis
+    np.put(to, np.arange(r_point, from_, -8), 1)
+    if np.all(bs_one * l_array == 0):
+        if np.all(l_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(l_array)
+    else:
+        max_dis = np.min(l_array[np.nonzero(l_array * bs_one)])
+    l_point = from_ - 8 * max_dis
+    np.put(to, np.arange(l_point, from_, 8), 1)
     return to
 
 
@@ -722,53 +923,65 @@ def _queen_moves(
     return r_move + b_move
 
 
+def _queen_effect(bs: np.ndarray, from_: int) -> np.ndarray:
+    r_ef = _rook_effect(bs, from_)
+    b_ef = _bishop_effect(bs, from_)
+    # r_moveとb_moveは共通項がないので足してよい
+    return r_ef + b_ef
+
+
 def _king_moves(bs: np.ndarray, from_: int, turn: int):
     to = np.zeros(64, dtype=np.int32)
     u, d, l_, r = _is_side(from_)
     if not u:
-        if _owner(bs[from_ + 1]) == turn:
-            to[from_ + 1] = 2
-        else:
+        if _owner(bs[from_ + 1]) != turn:
             to[from_ + 1] = 1
         if not l_:
-            if _owner(bs[from_ - 7]) == turn:
-                to[from_ - 7] = 2
-            else:
+            if _owner(bs[from_ - 7]) != turn:
                 to[from_ - 7] = 1
         if not r:
-            if _owner(bs[from_ + 9]) == turn:
-                to[from_ + 9] = 2
-            else:
+            if _owner(bs[from_ + 9]) != turn:
                 to[from_ + 9] = 1
     if not l_:
-        if _owner(bs[from_ - 8]) == turn:
-            to[from_ - 8] = 2
-        else:
+        if _owner(bs[from_ - 8]) != turn:
             to[from_ - 8] = 1
     if not r:
-        if _owner(bs[from_ + 8]) == turn:
-            to[from_ + 8] = 2
-        else:
+        if _owner(bs[from_ + 8]) != turn:
             to[from_ + 8] = 1
     if not d:
-        if _owner(bs[from_ - 1]) == turn:
-            to[from_ - 1] = 2
-        else:
+        if _owner(bs[from_ - 1]) != turn:
             to[from_ - 1] = 1
         if not l_:
-            if _owner(bs[from_ - 9]) == turn:
-                to[from_ - 9] = 2
-            else:
+            if _owner(bs[from_ - 9]) != turn:
                 to[from_ - 9] = 1
         if not r:
-            if _owner(bs[from_ + 7]) == turn:
-                to[from_ + 7] = 2
-            else:
+            if _owner(bs[from_ + 7]) != turn:
                 to[from_ + 7] = 1
     return to
 
 
-# 自分の駒へのmoveは2、その他の場所へのmoveは1として記録
+def _king_effect(from_: int):
+    to = np.zeros(64, dtype=np.int32)
+    u, d, l_, r = _is_side(from_)
+    if not u:
+        to[from_ + 1] = 1
+        if not l_:
+            to[from_ - 7] = 1
+        if not r:
+            to[from_ + 9] = 1
+    if not l_:
+        to[from_ - 8] = 1
+    if not r:
+        to[from_ + 8] = 1
+    if not d:
+        to[from_ - 1] = 1
+        if not l_:
+            to[from_ - 9] = 1
+        if not r:
+            to[from_ + 7] = 1
+    return to
+
+
 def _piece_moves(
     bs: np.ndarray, from_: int, piece: int, pins: np.ndarray
 ) -> np.ndarray:
@@ -791,6 +1004,25 @@ def _piece_moves(
         return _king_moves(bs, from_, turn)
 
 
+def _piece_effect(bs: np.ndarray, from_: int, piece: int) -> np.ndarray:
+    if piece == 0:
+        return np.zeros(64, dtype=np.int32)
+    turn = (piece - 1) // 6
+    p = piece % 6
+    if p == 1:
+        return _pawn_effects(from_, turn)
+    elif p == 2:
+        return _knight_effect(from_)
+    elif p == 3:
+        return _bishop_effect(bs, from_)
+    elif p == 4:
+        return _rook_effect(bs, from_)
+    elif p == 5:
+        return _queen_effect(bs, from_)
+    else:
+        return _king_effect(from_)
+
+
 def _create_actions(from_: int, to: np.ndarray) -> np.ndarray:
     actions = np.zeros(4608, dtype=np.int32)
     for i in range(64):
@@ -804,40 +1036,34 @@ def _create_actions(from_: int, to: np.ndarray) -> np.ndarray:
     return actions
 
 
+def _is_check(bs: np.ndarray, turn: int, kp: int) -> bool:
+    return _is_check2(bs, turn, kp)[0] != 0
+
+
+def _is_check2(bs: np.ndarray, turn: int, kp: int) -> Tuple[int, np.ndarray]:
+    checking_piece = np.zeros(64, dtype=np.int32)
+    e_color = (turn + 1) % 2
+    pe = _pawn_effects(kp, turn)
+    np.put(checking_piece, np.where(bs * pe == 1 + 6 * e_color), 1)
+    ke = _knight_effect(kp)
+    np.put(checking_piece, np.where(bs * ke == 2 + 6 * e_color), 1)
+    be = _bishop_effect(bs, kp)
+    np.put(checking_piece, np.where(bs * be == 3 + 6 * e_color), 1)
+    re = _rook_effect(bs, kp)
+    np.put(checking_piece, np.where(bs * re == 4 + 6 * e_color), 1)
+    qe = _queen_effect(bs, kp)
+    np.put(checking_piece, np.where(bs * qe == 5 + 6 * e_color), 1)
+    return np.count_nonzero(checking_piece), checking_piece
+
+
 def _effected_positions(bs: np.ndarray, turn: int) -> np.ndarray:
     effects = np.zeros(64, dtype=np.int32)
     for i in range(64):
         piece = bs[i]
-        if _owner(piece) != turn:
-            continue
-        if piece % 6 == 1:
-            effect = _pawn_effects(i, turn)
-        else:
-            effect = _piece_moves(bs, i, piece, np.zeros(64, dtype=np.int32))
-        effects += effect
+        effects += _piece_effect(bs, i, piece) if _owner(piece) == turn else 0
     return effects
 
 
-def _is_check(bs: np.ndarray, turn: int, kp: int) -> bool:
-    effects = _effected_positions(bs, (turn + 1) % 2)
-    return effects[kp] != 0
-
-
-# 王手している駒の位置も返すis_check
-def _is_check2(bs: np.ndarray, turn: int, kp: int) -> Tuple[int, np.ndarray]:
-    num_check = 0
-    checking_piece = np.zeros(64, dtype=np.int32)
-    for i in range(64):
-        piece = bs[i]
-        if _owner(piece) != turn and _owner(piece) != 0:
-            continue
-        moves = _piece_moves(bs, i, piece, np.zeros(64, dtype=np.int32))
-        if moves[kp] == 1:
-            num_check += 1
-            checking_piece[i] = 1
-    return num_check, checking_piece
-
-
 def _can_left_castling(state: ChessState, bs: np.ndarray) -> bool:
     if state.turn == 0:
         if state.wk_move_count or state.wr1_move_count:
@@ -955,20 +1181,12 @@ def _is_mate(state: ChessState, actions: np.ndarray) -> bool:
 # point1とpoint2の間の位置を返す
 # point2も含める
 def _between(point1: int, point2: int) -> np.ndarray:
-    flag = True
-    between = np.zeros(81, dtype=np.int32)
-    between[point2] = 1
+    between = np.zeros(64, dtype=np.int32)
     direction = _point_to_direction(point1, point2)
     if direction == -1:
         return between
     dif = _direction_to_dif(direction)
-    p = point1
-    for i in range(7):
-        p += dif
-        if p == point2:
-            flag = False
-        if flag:
-            between[p] = 1
+    np.put(between, np.arange(point1 + dif, point2 + dif, dif), 1)
     return between
 
 
@@ -982,9 +1200,8 @@ def _is_mate_check(state: ChessState, kp: int, pins: np.ndarray) -> bool:
     kingless_bs[kp] = 0
     kingless_effects = _effected_positions(kingless_bs, _another_color(state))
     # king_moveが1かつkingless_effectsが0の地点があれば詰みではない
-    for i in range(64):
-        if king_move[i] == 1 and kingless_effects[i] == 0:
-            return False
+    if np.any(king_move * (kingless_effects + 1) == 1):
+        return False
     # Kingを動かさない回避手の存在判定
     # 2枚以上からCheckされている場合→詰み
     if num_check == 2:
@@ -997,9 +1214,8 @@ def _is_mate_check(state: ChessState, kp: int, pins: np.ndarray) -> bool:
             continue
         pm = _piece_moves(bs, i, piece, pins)
         # checkしている駒やその駒との間に動ける駒がある場合、詰みでない
-        for j in range(64):
-            if bet[j] == 1 and pm[j] == 1:
-                return False
+        if np.any(bet * pm == 1):
+            return False
     return True
 
 
@@ -1011,9 +1227,8 @@ def _is_mate_non_check(state: ChessState, kp: int, pins: np.ndarray) -> bool:
     kingless_bs[kp] = 0
     kingless_effects = _effected_positions(kingless_bs, _another_color(state))
     # king_moveが1かつkingless_effectsが0の地点があれば詰みではない
-    for i in range(64):
-        if king_move[i] == 1 and kingless_effects[i] == 0:
-            return False
+    if np.any(king_move * (kingless_effects + 1) == 1):
+        return False
     # Kingが逃げる手以外の合法手を調べる
     for i in range(64):
         piece = kingless_bs[i]
@@ -1063,274 +1278,47 @@ def _is_legal_action(state: ChessState, action: int, pins: np.ndarray):
     return p_actions[format_action] == 1
 
 
-def _up_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
+def _direction_to_pin(direction: int) -> int:
+    if direction == 0 or direction == 1:
+        return 1
+    if direction == 2 or direction == 3:
+        return 2
+    if direction == 4 or direction == 5:
+        return 3
+    if direction == 6 or direction == 7:
+        return 4
+    return 0
+
+
+def _direction_pin(
+    bs: np.ndarray,
+    turn: int,
+    king_point: int,
+    direction: int,
+    array: np.ndarray,
 ) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    # 上方向のピン
-    u = king_point
-    u_num = 0
-    u_piece = -1
-    u_flag = False
-    # 自分より上は最大7マスしかない
-    for i in range(7):
-        # 探索位置の更新
-        u += 1
-        # 探索が終わっている場合はスルー
-        if u_flag:
-            continue
-        # 盤外や同じ列にない場合は弾く
-        if not _is_in_board(u) or not _is_same_column(king_point, u):
-            continue
-        piece = bs[u]
-        # 駒がない場合は処理をしない
-        if piece == 0:
-            continue
-        # 駒がある場合
-        u_num += 1
-        # 1枚目
-        if u_num == 1:
-            # 自分の駒なら位置を記録
-            if _owner(piece) == turn:
-                u_piece = u
-            # 相手の駒なら関係ないので探索終了
-            else:
-                u_flag = True
-        # 2枚目
-        if u_num == 2:
-            # Rook, Queenによってピン
-            if turn == 0 and (piece == 10 or piece == 11):
-                new_array[u_piece] = 1
-            elif turn == 1 and (piece == 4 or piece == 5):
-                new_array[u_piece] = 1
-            # 一回通ったら探索する必要はない
-            u_flag = True
-    return new_array
-
-
-def _down_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    d = king_point
-    d_num = 0
-    d_piece = -1
-    d_flag = False
-    for i in range(7):
-        d -= 1
-        if d_flag:
-            continue
-        if not _is_in_board(d) or not _is_same_column(king_point, d):
-            continue
-        piece = bs[d]
-        if piece == 0:
-            continue
-        d_num += 1
-        if d_num == 1:
-            if _owner(piece) == turn:
-                d_piece = d
-            else:
-                d_flag = True
-        if d_num == 2:
-            if turn == 0 and (piece == 10 or piece == 11):
-                new_array[d_piece] = 1
-            elif turn == 1 and (piece == 4 or piece == 5):
-                new_array[d_piece] = 1
-            d_flag = True
-    return new_array
-
-
-def _left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    l_ = king_point
-    l_num = 0
-    l_piece = -1
-    l_flag = False
-    for i in range(7):
-        l_ -= 8
-        if l_flag:
-            continue
-        if not _is_in_board(l_) or not _is_same_row(king_point, l_):
-            continue
-        piece = bs[l_]
-        if piece == 0:
-            continue
-        l_num += 1
-        if l_num == 1:
-            if _owner(piece) == turn:
-                l_piece = l_
-            else:
-                l_flag = True
-        if l_num == 2:
-            if turn == 0 and (piece == 10 or piece == 11):
-                new_array[l_piece] = 2
-            elif turn == 1 and (piece == 4 or piece == 5):
-                new_array[l_piece] = 2
-            l_flag = True
-    return new_array
-
-
-def _right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    r = king_point
-    r_num = 0
-    r_piece = -1
-    r_flag = False
-    for i in range(7):
-        r += 8
-        if r_flag:
-            continue
-        if not _is_in_board(r) or not _is_same_row(king_point, r):
-            continue
-        piece = bs[r]
-        if piece == 0:
-            continue
-        r_num += 1
-        if r_num == 1:
-            if _owner(piece) == turn:
-                r_piece = r
-            else:
-                r_flag = True
-        if r_num == 2:
-            if turn == 0 and (piece == 10 or piece == 11):
-                new_array[r_piece] = 2
-            elif turn == 1 and (piece == 4 or piece == 5):
-                new_array[r_piece] = 2
-            r_flag = True
-    return new_array
-
-
-def _up_right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    ur = king_point
-    ur_num = 0
-    ur_piece = -1
-    ur_flag = False
-    for i in range(7):
-        ur += 9
-        if ur_flag:
-            continue
-        if not _is_in_board(ur) or not _is_same_rising(king_point, ur):
-            continue
-        piece = bs[ur]
-        if piece == 0:
-            continue
-        ur_num += 1
-        if ur_num == 1:
-            if _owner(piece) == turn:
-                ur_piece = ur
-            else:
-                ur_flag = True
-        # 2枚目
-        if ur_num == 2:
-            # Bishop, Queenによってピン
-            if turn == 0 and (piece == 9 or piece == 11):
-                new_array[ur_piece] = 3
-            elif turn == 1 and (piece == 3 or piece == 5):
-                new_array[ur_piece] = 3
-            ur_flag = True
-    return new_array
-
-
-def _down_left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    dl = king_point
-    dl_num = 0
-    dl_piece = -1
-    dl_flag = False
-    for i in range(7):
-        dl -= 9
-        if dl_flag:
-            continue
-        if not _is_in_board(dl) or not _is_same_rising(king_point, dl):
-            continue
-        piece = bs[dl]
-        if piece == 0:
-            continue
-        dl_num += 1
-        if dl_num == 1:
-            if _owner(piece) == turn:
-                dl_piece = dl
-            else:
-                dl_flag = True
-        if dl_num == 2:
-            if turn == 0 and (piece == 9 or piece == 11):
-                new_array[dl_piece] = 3
-            elif turn == 1 and (piece == 3 or piece == 5):
-                new_array[dl_piece] = 3
-            dl_flag = True
-    return new_array
-
-
-def _up_left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    ul = king_point
-    ul_num = 0
-    ul_piece = -1
-    ul_flag = False
-    for i in range(7):
-        ul -= 7
-        if ul_flag:
-            continue
-        if not _is_in_board(ul) or not _is_same_row(king_point, ul):
-            continue
-        piece = bs[ul]
-        if piece == 0:
-            continue
-        ul_num += 1
-        if ul_num == 1:
-            if _owner(piece) == turn:
-                ul_piece = ul
-            else:
-                ul_flag = True
-        if ul_num == 2:
-            if turn == 0 and (piece == 9 or piece == 11):
-                new_array[ul_piece] = 4
-            elif turn == 1 and (piece == 3 or piece == 5):
-                new_array[ul_piece] = 4
-            ul_flag = True
-    return new_array
-
-
-def _down_right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    dr = king_point
-    dr_num = 0
-    dr_piece = -1
-    dr_flag = False
-    for i in range(7):
-        dr += 7
-        if dr_flag:
-            continue
-        if not _is_in_board(dr) or not _is_same_row(king_point, dr):
-            continue
-        piece = bs[dr]
-        if piece == 0:
-            continue
-        dr_num += 1
-        if dr_num == 1:
-            if _owner(piece) == turn:
-                dr_piece = dr
-            else:
-                dr_flag = True
-        if dr_num == 2:
-            if turn == 0 and (piece == 9 or piece == 11):
-                new_array[dr_piece] = 4
-            elif turn == 1 and (piece == 3 or piece == 5):
-                new_array[dr_piece] = 4
-            dr_flag = True
+    new_array = array
+    dir_array = _dis_direction_array(king_point, direction)
+    e_turn = (turn + 1) % 2
+    bs_one = np.where(bs == 0, 0, 1)
+    dir_one_array = dir_array * bs_one
+    if np.count_nonzero(dir_one_array) <= 1:
+        return new_array
+    dif = _direction_to_dif(direction)
+    dis1 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 0)[0]
+    dis2 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 1)[1]
+    point1 = king_point + dis1 * dif
+    point2 = king_point + dis2 * dif
+    piece1 = bs[point1]
+    piece2 = bs[point2]
+    if _owner(piece1) != turn:
+        return new_array
+    if direction <= 3:
+        if piece2 == 4 + 6 * e_turn or piece2 == 5 + 6 * e_turn:
+            new_array[point1] = _direction_to_pin(direction)
+    else:
+        if piece2 == 3 + 6 * e_turn or piece2 == 5 + 6 * e_turn:
+            new_array[point1] = _direction_to_pin(direction)
     return new_array
 
 
@@ -1338,12 +1326,6 @@ def _pin(state: ChessState, kp: int):
     bs = _board_status(state)
     turn = state.turn
     pins = np.zeros(64, dtype=np.int32)
-    pins = _up_pin(bs, turn, kp, pins)
-    pins = _up_left_pin(bs, turn, kp, pins)
-    pins = _up_right_pin(bs, turn, kp, pins)
-    pins = _left_pin(bs, turn, kp, pins)
-    pins = _right_pin(bs, turn, kp, pins)
-    pins = _down_pin(bs, turn, kp, pins)
-    pins = _down_left_pin(bs, turn, kp, pins)
-    pins = _down_right_pin(bs, turn, kp, pins)
+    for i in range(8):
+        pins = _direction_pin(bs, turn, kp, i, pins)
     return pins
