@@ -44,7 +44,7 @@ class ShogiState:
 
 def init() -> ShogiState:
     board = _make_init_board()
-    state = ShogiState(board=board)
+    state = ShogiState(board=board, hand=np.zeros(14, dtype=np.int32))
     return _init_legal_actions(state)
 
 
@@ -218,13 +218,80 @@ def _gold_move(turn: int) -> np.ndarray:
 
 def _king_move() -> np.ndarray:
     array = np.zeros((9, 9), dtype=np.int32)
-    for i in range(3):
-        for j in range(3):
-            array[3 + i][3 + j] = 1
+    array[3:6, 3:6] = 1
     array[4][4] = 0
     return array
 
 
+def _is_side(point: int) -> Tuple[bool, bool, bool, bool]:
+    is_up = point % 9 == 0
+    is_down = point % 9 == 8
+    is_left = point >= 72
+    is_right = point <= 8
+    return is_up, is_down, is_left, is_right
+
+
+# 桂馬用
+def _is_second_line(point: int) -> Tuple[bool, bool]:
+    u = point % 9 <= 1
+    d = point % 9 >= 7
+    return u, d
+
+
+# point(0~80)を座標((0, 0)~(8, 8))に変換
+def _point_to_location(point: int) -> Tuple[int, int]:
+    return point // 9, point % 9
+
+
+def _cut_outside(array: np.ndarray, point: int) -> np.ndarray:
+    new_array = array
+    u, d, l, r = _is_side(point)
+    u2, d2 = _is_second_line(point)
+    # (4, 4)での動きを基準にはみ出すところをカットする
+    if u:
+        new_array[:, 3] *= 0
+    if d:
+        new_array[:, 5] *= 0
+    if r:
+        new_array[3, :] *= 0
+    if l:
+        new_array[5, :] *= 0
+    if u2:
+        new_array[:, 2] *= 0
+    if d2:
+        new_array[:, 6] *= 0
+    return new_array
+
+
+def _action_board(array: np.ndarray, point: int) -> np.ndarray:
+    new_array = array
+    y, t = _point_to_location(point)
+    new_array = _cut_outside(new_array, point)
+    return np.roll(new_array, (y - 4, t - 4), axis=(0, 1)).reshape(81)
+
+
+POINT_MOVES = np.zeros((81, 29, 81), dtype=np.int32)
+for i in range(81):
+    POINT_MOVES[i][1] = _action_board(_pawn_move(0), i)
+    POINT_MOVES[i][3] = _action_board(_knight_move(0), i)
+    POINT_MOVES[i][4] = _action_board(_silver_move(0), i)
+    POINT_MOVES[i][7] = _action_board(_gold_move(0), i)
+    POINT_MOVES[i][8] = _action_board(_king_move(), i)
+    POINT_MOVES[i][9] = POINT_MOVES[i][7]
+    POINT_MOVES[i][10] = POINT_MOVES[i][7]
+    POINT_MOVES[i][11] = POINT_MOVES[i][7]
+    POINT_MOVES[i][12] = POINT_MOVES[i][7]
+    POINT_MOVES[i][15] = _action_board(_pawn_move(1), i)
+    POINT_MOVES[i][17] = _action_board(_knight_move(1), i)
+    POINT_MOVES[i][18] = _action_board(_silver_move(1), i)
+    POINT_MOVES[i][21] = _action_board(_gold_move(1), i)
+    POINT_MOVES[i][22] = _action_board(_king_move(), i)
+    POINT_MOVES[i][23] = POINT_MOVES[i][21]
+    POINT_MOVES[i][24] = POINT_MOVES[i][21]
+    POINT_MOVES[i][25] = POINT_MOVES[i][21]
+    POINT_MOVES[i][26] = POINT_MOVES[i][21]
+
+
 # dlshogiのactionはdirection(動きの方向)とto（駒の処理後の座標）に依存
 def _dlshogi_action(direction: int, to: int) -> int:
     return direction * 81 + to
@@ -250,6 +317,50 @@ def _is_same_declining(_from: int, to: int) -> bool:
     return _from // 9 + _from % 9 == to // 9 + to % 9
 
 
+def _is_same_line(from_: int, to: int, direction: int):
+    dir = direction % 10
+    if dir == 0 or dir == 5:
+        return _is_same_column(from_, to)
+    elif dir == 3 or dir == 4:
+        return _is_same_row(from_, to)
+    elif dir == 2 or dir == 6:
+        return _is_same_rising(from_, to)
+    elif dir == 1 or dir == 7:
+        return _is_same_declining(from_, to)
+    else:
+        return False
+
+
+def _dis_direction_array(from_: int, turn: int, direction: int):
+    array = np.zeros(81, dtype=np.int32)
+    dif = _direction_to_dif(direction, turn)
+    to = from_ + dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 1
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 2
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 3
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 4
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 5
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 6
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 7
+    to += dif
+    if _is_in_board(to) and _is_same_line(from_, to, direction):
+        array[to] = 8
+    return array
+
+
 # fromの座標とtoの座標からdirを生成
 def _point_to_direction(_from: int, to: int, promote: bool, turn: int) -> int:
     direction = -1
@@ -353,7 +464,7 @@ def _direction_to_from(
     _from = -1
     for i in range(9):
         f -= dif
-        if 80 >= f >= 0 and _from == -1 and _piece_type(state, f) != 0:
+        if _is_in_board(f) and _from == -1 and _piece_type(state, f) != 0:
             _from = f
     if direction >= 10:
         return _from, True
@@ -432,10 +543,7 @@ def _owner(piece: int) -> int:
 # 盤面のどこに何の駒があるかをnp.arrayに移したもの
 # 同じ座標に複数回piece_typeを使用する場合はこちらを使った方が良い
 def _board_status(state: ShogiState) -> np.ndarray:
-    board = np.zeros(81, dtype=np.int32)
-    for i in range(81):
-        board[i] = _piece_type(state, i)
-    return board
+    return state.board.argmax(axis=0)
 
 
 # 駒の持ち主の判定
@@ -452,7 +560,7 @@ def _move(
     state: ShogiState,
     action: ShogiAction,
 ) -> ShogiState:
-    s = copy.deepcopy(state)
+    s = state
     s.board[action.piece][action.from_] = 0
     s.board[0][action.from_] = 1
     s.board[action.captured][action.to] = 0
@@ -466,225 +574,203 @@ def _move(
 
 
 def _drop(state: ShogiState, action: ShogiAction) -> ShogiState:
-    s = copy.deepcopy(state)
+    s = state
     s.hand[_piece_to_hand(action.piece)] -= 1
     s.board[action.piece][action.to] = 1
     s.board[0][action.to] = 0
     return s
 
 
-def _is_side(point: int) -> Tuple[bool, bool, bool, bool]:
-    is_up = point % 9 == 0
-    is_down = point % 9 == 8
-    is_left = point >= 72
-    is_right = point <= 8
-    return is_up, is_down, is_left, is_right
-
-
-# 桂馬用
-def _is_second_line(point: int) -> Tuple[bool, bool]:
-    u = point % 9 <= 1
-    d = point % 9 >= 7
-    return u, d
-
-
-# point(0~80)を座標((0, 0)~(8, 8))に変換
-def _point_to_location(point: int) -> Tuple[int, int]:
-    return point // 9, point % 9
-
-
-def _cut_outside(array: np.ndarray, point: int) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    u, d, l, r = _is_side(point)
-    u2, d2 = _is_second_line(point)
-    # (4, 4)での動きを基準にはみ出すところをカットする
-    if u:
-        new_array[:, 3] *= 0
-    if d:
-        new_array[:, 5] *= 0
-    if r:
-        new_array[3, :] *= 0
-    if l:
-        new_array[5, :] *= 0
-    if u2:
-        new_array[:, 2] *= 0
-    if d2:
-        new_array[:, 6] *= 0
-    return new_array
-
-
-def _action_board(array: np.ndarray, point: int) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    y, t = _point_to_location(point)
-    new_array = _cut_outside(new_array, point)
-    return np.roll(new_array, (y - 4, t - 4), axis=(0, 1)).reshape(81)
-
-
-def _lance_move(bs: np.ndarray, point: int, turn: int) -> np.ndarray:
-    array = np.zeros(81, dtype=np.int32)
-    flag = True
-    for i in range(8):
-        if turn == 0:
-            p = point - 1 - i
+def _lance_move(bs: np.ndarray, from_: int, turn: int) -> np.ndarray:
+    to = np.zeros(81, dtype=np.int32)
+    u_array = _dis_direction_array(from_, turn, 0)
+    bs_one = np.where(bs == 0, 0, 1)
+    if np.all(u_array * bs_one == 0):
+        if np.all(u_array == 0):
+            max_dis = 0
         else:
-            p = point + 1 + i
-        # pが盤内、かつ同じ列、かつフラグが折れていない場合は利きが通っている
-        if _is_in_board(p) and _is_same_column(point, p) and flag:
-            array[p] = 1
-        # pが盤外、または空白でない場合はフラグを折ってそれより奥に到達できないようにする
-        if not _is_in_board(p) or bs[p] != 0:
-            flag = False
-    return array
+            max_dis = np.max(u_array)
+    else:
+        max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
+    if turn == 0:
+        u_point = from_ - max_dis
+        np.put(to, np.arange(u_point, from_, 1), 1)
+    else:
+        d_point = from_ + max_dis
+        np.put(to, np.arange(d_point, from_, -1), 1)
+    return to
 
 
-def _bishop_move(bs: np.ndarray, point: int) -> np.ndarray:
-    array = np.zeros(81, dtype=np.int32)
-    ur_flag = True
-    ul_flag = True
-    dr_flag = True
-    dl_flag = True
-    for i in range(8):
-        ur = point - 10 * (1 + i)
-        ul = point + 8 * (1 + i)
-        dr = point - 8 * (1 + i)
-        dl = point + 10 * (1 + i)
-        if _is_in_board(ur) and _is_same_rising(point, ur) and ur_flag:
-            array[ur] = 1
-        if _is_in_board(ul) and _is_same_declining(point, ul) and ul_flag:
-            array[ul] = 1
-        if _is_in_board(dr) and _is_same_declining(point, dr) and dr_flag:
-            array[dr] = 1
-        if _is_in_board(dl) and _is_same_rising(point, dl) and dl_flag:
-            array[dl] = 1
-        if not _is_in_board(ur) or bs[ur] != 0:
-            ur_flag = False
-        if not _is_in_board(ul) or bs[ul] != 0:
-            ul_flag = False
-        if not _is_in_board(dr) or bs[dr] != 0:
-            dr_flag = False
-        if not _is_in_board(dl) or bs[dl] != 0:
-            dl_flag = False
-    return array
+def _bishop_move(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = np.zeros(81, dtype=np.int32)
+    ur_array = _dis_direction_array(from_, 0, 2)
+    ul_array = _dis_direction_array(from_, 0, 1)
+    dr_array = _dis_direction_array(from_, 0, 7)
+    dl_array = _dis_direction_array(from_, 0, 6)
+    bs_one = np.where(bs == 0, 0, 1)
+    if np.all(bs_one * ur_array == 0):
+        if np.all(ur_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(ur_array)
+    else:
+        max_dis = np.min(ur_array[np.nonzero(ur_array * bs_one)])
+    ur_point = from_ - 10 * max_dis
+    np.put(to, np.arange(ur_point, from_, 10), 1)
+    if np.all(bs_one * ul_array == 0):
+        if np.all(ul_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(ul_array)
+    else:
+        max_dis = np.min(ul_array[np.nonzero(ul_array * bs_one)])
+    ul_point = from_ + 8 * max_dis
+    np.put(to, np.arange(ul_point, from_, -8), 1)
+    if np.all(bs_one * dr_array == 0):
+        if np.all(dr_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(dr_array)
+    else:
+        max_dis = np.min(dr_array[np.nonzero(dr_array * bs_one)])
+    dr_point = from_ - 8 * max_dis
+    np.put(to, np.arange(dr_point, from_, 8), 1)
+    if np.all(bs_one * dl_array == 0):
+        if np.all(dl_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(dl_array)
+    else:
+        max_dis = np.min(dl_array[np.nonzero(dl_array * bs_one)])
+    dl_point = from_ + 10 * max_dis
+    np.put(to, np.arange(dl_point, from_, -10), 1)
+    return to
+    # for i in range(8):
+    #    ur = point - 10 * (1 + i)
+    #    ul = point + 8 * (1 + i)
+    #    dr = point - 8 * (1 + i)
+    #    dl = point + 10 * (1 + i)
+    #    if _is_in_board(ur) and _is_same_rising(point, ur) and ur_flag:
+    #        array[ur] = 1
+    #    if _is_in_board(ul) and _is_same_declining(point, ul) and ul_flag:
+    #        array[ul] = 1
+    #    if _is_in_board(dr) and _is_same_declining(point, dr) and dr_flag:
+    #        array[dr] = 1
+    #    if _is_in_board(dl) and _is_same_rising(point, dl) and dl_flag:
+    #        array[dl] = 1
+    #    if not _is_in_board(ur) or bs[ur] != 0:
+    #        ur_flag = False
+    #   if not _is_in_board(ul) or bs[ul] != 0:
+    #        ul_flag = False
+    #    if not _is_in_board(dr) or bs[dr] != 0:
+    #        dr_flag = False
+    #    if not _is_in_board(dl) or bs[dl] != 0:
+    #        dl_flag = False
+    # return array
 
 
-def _rook_move(bs: np.ndarray, point: int) -> np.ndarray:
-    array = np.zeros(81, dtype=np.int32)
-    u_flag = True
-    d_flag = True
-    r_flag = True
-    l_flag = True
-    for i in range(8):
-        u = point - 1 * (1 + i)
-        d = point + 1 * (1 + i)
-        r = point - 9 * (1 + i)
-        l_ = point + 9 * (1 + i)
-        if _is_in_board(u) and _is_same_column(point, u) and u_flag:
-            array[u] = 1
-        if _is_in_board(d) and _is_same_column(point, d) and d_flag:
-            array[d] = 1
-        if _is_in_board(r) and _is_same_row(point, r) and r_flag:
-            array[r] = 1
-        if _is_in_board(l_) and _is_same_row(point, l_) and l_flag:
-            array[l_] = 1
-        if not _is_in_board(u) or bs[u] != 0:
-            u_flag = False
-        if not _is_in_board(d) or bs[d] != 0:
-            d_flag = False
-        if not _is_in_board(r) or bs[r] != 0:
-            r_flag = False
-        if not _is_in_board(l_) or bs[l_] != 0:
-            l_flag = False
-    return array
+def _rook_move(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = np.zeros(81, dtype=np.int32)
+    u_array = _dis_direction_array(from_, 0, 0)
+    d_array = _dis_direction_array(from_, 0, 5)
+    r_array = _dis_direction_array(from_, 0, 4)
+    l_array = _dis_direction_array(from_, 0, 3)
+    bs_one = np.where(bs == 0, 0, 1)
+    if np.all(bs_one * u_array == 0):
+        if np.all(u_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(u_array)
+    else:
+        max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
+    ur_point = from_ - max_dis
+    np.put(to, np.arange(ur_point, from_, 1), 1)
+    if np.all(bs_one * d_array == 0):
+        if np.all(d_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(d_array)
+    else:
+        max_dis = np.min(d_array[np.nonzero(d_array * bs_one)])
+    ul_point = from_ + max_dis
+    np.put(to, np.arange(ul_point, from_, -1), 1)
+    if np.all(bs_one * r_array == 0):
+        if np.all(r_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(r_array)
+    else:
+        max_dis = np.min(r_array[np.nonzero(r_array * bs_one)])
+    r_point = from_ - 9 * max_dis
+    np.put(to, np.arange(r_point, from_, 9), 1)
+    if np.all(bs_one * l_array == 0):
+        if np.all(l_array == 0):
+            max_dis = 0
+        else:
+            max_dis = np.max(l_array)
+    else:
+        max_dis = np.min(l_array[np.nonzero(l_array * bs_one)])
+    l_point = from_ + 9 * max_dis
+    np.put(to, np.arange(l_point, from_, -9), 1)
+    return to
 
 
-def _horse_move(bs: np.ndarray, point: int) -> np.ndarray:
-    array = _bishop_move(bs, point)
-    u, d, l, r = _is_side(point)
+def _horse_move(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = _bishop_move(bs, from_)
+    u, d, l, r = _is_side(from_)
     if not u:
-        array[point - 1] = 1
+        to[from_ - 1] = 1
     if not d:
-        array[point + 1] = 1
+        to[from_ + 1] = 1
     if not r:
-        array[point - 9] = 1
+        to[from_ - 9] = 1
     if not l:
-        array[point + 9] = 1
-    return array
+        to[from_ + 9] = 1
+    return to
 
 
-def _dragon_move(bs: np.ndarray, point: int) -> np.ndarray:
-    array = _rook_move(bs, point)
-    u, d, l, r = _is_side(point)
+def _dragon_move(bs: np.ndarray, from_: int) -> np.ndarray:
+    to = _rook_move(bs, from_)
+    u, d, l, r = _is_side(from_)
     if not u:
         if not r:
-            array[point - 10] = 1
+            to[from_ - 10] = 1
         if not l:
-            array[point + 8] = 1
+            to[from_ + 8] = 1
     if not d:
         if not r:
-            array[point - 8] = 1
+            to[from_ - 8] = 1
         if not l:
-            array[point + 10] = 1
-    return array
+            to[from_ + 10] = 1
+    return to
 
 
 # 駒種と位置から到達できる場所を返す
 def _piece_moves(bs: np.ndarray, piece: int, point: int) -> np.ndarray:
-    turn = _owner(piece)
-    piece_type = piece % 14
-    # 歩の動き
-    if piece_type == 1:
-        return _action_board(_pawn_move(turn), point)
+    moves = POINT_MOVES[point][piece]
     # 香車の動き
-    if piece_type == 2:
-        return _lance_move(bs, point, turn)
-    # 桂馬の動き
-    if piece_type == 3:
-        return _action_board(_knight_move(turn), point)
-    # 銀の動き
-    if piece_type == 4:
-        return _action_board(_silver_move(turn), point)
+    if piece == 2:
+        moves = _lance_move(bs, point, 0)
+    if piece == 16:
+        moves = _lance_move(bs, point, 1)
     # 角の動き
-    if piece_type == 5:
-        return _bishop_move(bs, point)
+    if piece == 5 or piece == 19:
+        moves = _bishop_move(bs, point)
     # 飛車の動き
-    if piece_type == 6:
-        return _rook_move(bs, point)
-    # 金および成金の動き
-    if piece_type == 7 or 9 <= piece_type <= 12:
-        return _action_board(_gold_move(turn), point)
-    # 玉の動き
-    if piece_type == 8:
-        return _action_board(_king_move(), point)
+    if piece == 6 or piece == 20:
+        moves = _rook_move(bs, point)
     # 馬の動き
-    if piece_type == 13:
-        return _horse_move(bs, point)
+    if piece == 13 or piece == 27:
+        moves = _horse_move(bs, point)
     # 龍の動き
-    # piece_typeが0になってしまっているのでpieceで判別
     if piece == 14 or piece == 28:
-        return _dragon_move(bs, point)
-    return np.zeros(81, dtype=np.int32)
+        moves = _dragon_move(bs, point)
+    return moves
 
 
 # 小駒のactionのみを返すpiece_moves
 def _small_piece_moves(piece: int, point: int) -> np.ndarray:
-    turn = _owner(piece)
-    piece_type = piece % 14
-    # 歩の動き
-    if piece_type == 1:
-        return _action_board(_pawn_move(turn), point)
-    # 桂馬の動き
-    if piece_type == 3:
-        return _action_board(_knight_move(turn), point)
-    # 銀の動き
-    if piece_type == 4:
-        return _action_board(_silver_move(turn), point)
-    # 金および成金の動き
-    if piece_type == 7 or 9 <= piece_type <= 12:
-        return _action_board(_gold_move(turn), point)
-    # 玉の動き
-    if piece_type == 8:
-        return _action_board(_king_move(), point)
-    return np.zeros(81, dtype=np.int32)
+    return POINT_MOVES[point][piece]
 
 
 # 敵陣かどうか
@@ -721,76 +807,55 @@ def _create_actions(piece: int, _from: int, to: np.ndarray) -> np.ndarray:
 
 
 def _create_piece_actions(piece: int, _from: int) -> np.ndarray:
-    moves = _small_piece_moves(piece, _from)
-    return _create_actions(piece, _from, moves)
+    return _create_actions(piece, _from, POINT_MOVES[_from][piece])
 
 
 # actionを追加する
 def _add_action(add_array: np.ndarray, origin_array: np.ndarray) -> np.ndarray:
-    new_array = copy.deepcopy(origin_array)
-    for i in range(2754):
-        if add_array[i] == 1:
-            new_array[i] = 1
-    return new_array
+    return np.where(add_array == 1, 1, origin_array)
 
 
 # actionを削除する
 def _filter_action(
     filter_array: np.ndarray, origin_array: np.ndarray
 ) -> np.ndarray:
-    new_array = copy.deepcopy(origin_array)
-    for i in range(2754):
-        if filter_array[i] == 1:
-            new_array[i] = 0
-    return new_array
+    return np.where(filter_array == 1, 0, origin_array)
 
 
 # 駒の種類と位置から生成できるactionのフラグを立てる
 def _add_move_actions(piece: int, _from: int, array: np.ndarray) -> np.ndarray:
-    new_array = copy.deepcopy(array)
     actions = _create_piece_actions(piece, _from)
-    for i in range(2754):
-        if actions[i] == 1:
-            new_array[i] = 1
-    return new_array
+    return np.where(actions == 1, 1, array)
 
 
 # 駒の種類と位置から生成できるactionのフラグを折る
 def _filter_move_actions(
     piece: int, _from: int, array: np.ndarray
 ) -> np.ndarray:
-    new_array = copy.deepcopy(array)
     actions = _create_piece_actions(piece, _from)
-    for i in range(2754):
-        if actions[i] == 1:
-            new_array[i] = 0
-    return new_array
+    return np.where(actions == 1, 0, array)
 
 
 # 駒打ちのactionを追加する
 def _add_drop_actions(piece: int, array: np.ndarray) -> np.ndarray:
-    new_array = copy.deepcopy(array)
+    new_array = array
     direction = _hand_to_direction(piece)
-    for i in range(81):
-        action = _dlshogi_action(direction, i)
-        new_array[action] = 1
+    np.put(new_array, np.arange(81 * direction, 81 * (direction + 1)), 1)
     return new_array
 
 
 # 駒打ちのactionのフラグを折る
 def _filter_drop_actions(piece: int, array: np.ndarray) -> np.ndarray:
-    new_array = copy.deepcopy(array)
+    new_array = array
     direction = _hand_to_direction(piece)
-    for i in range(81):
-        action = _dlshogi_action(direction, i)
-        new_array[action] = 0
+    np.put(new_array, np.arange(81 * direction, 81 * (direction + 1)), 0)
     return new_array
 
 
 # stateからblack,white両方のlegal_actionsを生成する
 # 普段は使わないがlegal_actionsが設定されていない場合に使用
 def _init_legal_actions(state: ShogiState) -> ShogiState:
-    s = copy.deepcopy(state)
+    s = state
     bs = _board_status(s)
     # 移動の追加
     for i in range(81):
@@ -822,7 +887,7 @@ def _init_legal_actions(state: ShogiState) -> ShogiState:
 def _update_legal_move_actions(
     state: ShogiState, action: ShogiAction
 ) -> ShogiState:
-    s = copy.deepcopy(state)
+    s = state
     if s.turn == 0:
         player_actions = s.legal_actions_black
         enemy_actions = s.legal_actions_white
@@ -864,7 +929,7 @@ def _update_legal_move_actions(
 def _update_legal_drop_actions(
     state: ShogiState, action: ShogiAction
 ) -> ShogiState:
-    s = copy.deepcopy(state)
+    s = state
     if s.turn == 0:
         player_actions = s.legal_actions_black
     else:
@@ -889,12 +954,11 @@ def _update_legal_drop_actions(
 def _filter_my_piece_move_actions(
     turn: int, owner: np.ndarray, array: np.ndarray
 ) -> np.ndarray:
-    new_array = copy.deepcopy(array)
+    new_array = array
     for i in range(81):
         if owner[i] != turn:
             continue
-        for j in range(20):
-            new_array[81 * j + i] = 0
+        np.put(new_array, np.arange(i, 1620 + i, 81), 0)
     return new_array
 
 
@@ -902,12 +966,15 @@ def _filter_my_piece_move_actions(
 def _filter_occupied_drop_actions(
     turn: int, owner: np.ndarray, array: np.ndarray
 ) -> np.ndarray:
-    new_array = copy.deepcopy(array)
+    new_array = array
     for i in range(81):
         if owner[i] == 2:
             continue
-        for j in range(7):
-            new_array[81 * (j + 20 + 7 * turn) + i] = 0
+        np.put(
+            new_array,
+            np.arange(81 * (20 + 7 * turn) + i, 81 * (27 + 7 * turn) + i, 81),
+            0,
+        )
     return new_array
 
 
@@ -916,39 +983,18 @@ def _filter_occupied_drop_actions(
 # 自殺手や反則手はここでは除かない
 def _legal_actions(state: ShogiState) -> np.ndarray:
     if state.turn == 0:
-        action_array = copy.deepcopy(state.legal_actions_black)
+        action_array = state.legal_actions_black
     else:
-        action_array = copy.deepcopy(state.legal_actions_white)
+        action_array = state.legal_actions_white
     bs = _board_status(state)
     own = _pieces_owner(state)
     for i in range(81):
-        piece = _piece_type(state, i)
-        # 香車の動きを追加
-        if piece == 2 + 14 * state.turn:
+        piece = bs[i]
+        if _owner(piece) == state.turn:
             action_array = _add_action(
-                _create_actions(piece, i, _lance_move(bs, i, state.turn)),
+                _create_actions(piece, i, _piece_moves(bs, piece, i)),
                 action_array,
             )
-        # 角の動きを追加
-        if piece == 5 + 14 * state.turn:
-            action_array = _add_action(
-                _create_actions(piece, i, _bishop_move(bs, i)), action_array
-            )
-        # 飛車の動きを追加
-        if piece == 6 + 14 * state.turn:
-            action_array = _add_action(
-                _create_actions(piece, i, _rook_move(bs, i)), action_array
-            )
-        # 馬の動きを追加
-        if piece == 13 + 14 * state.turn:
-            action_array = _add_action(
-                _create_actions(piece, i, _horse_move(bs, i)), action_array
-            )
-        # 龍の動きを追加
-        if piece == 14 + 14 * state.turn:
-            action_array = _add_action(
-                _create_actions(piece, i, _dragon_move(bs, i)), action_array
-            )
     # 自分の駒がある位置への移動actionを除く
     action_array = _filter_my_piece_move_actions(state.turn, own, action_array)
     # 駒がある地点への駒打ちactionを除く
@@ -990,10 +1036,9 @@ def _is_double_pawn(state: ShogiState) -> bool:
     is_double_pawn = False
     bs = _board_status(state)
     for i in range(9):
-        num_pawn = 0
-        for j in range(9):
-            if bs[9 * i + j] == 1 + state.turn * 14:
-                num_pawn += 1
+        num_pawn = np.count_nonzero(
+            bs[9 * i : 9 * (i + 1)] == 1 + state.turn * 14
+        )
         if num_pawn >= 2:
             is_double_pawn = True
     return is_double_pawn
@@ -1003,301 +1048,77 @@ def _is_double_pawn(state: ShogiState) -> bool:
 def _is_stuck(state: ShogiState) -> bool:
     is_stuck = False
     bs = _board_status(state)
-    for i in range(9):
-        for j in range(9):
-            piece = bs[9 * i + j]
-            if (
-                state.turn == 0
-                and (piece == 1 or piece == 2 or piece == 3)
-                and j == 0
-            ):
-                is_stuck = True
-            if state.turn == 0 and piece == 3 and j == 1:
-                is_stuck = True
-            if (
-                state.turn == 1
-                and (piece == 15 or piece == 16 or piece == 17)
-                and j == 8
-            ):
-                is_stuck = True
-            if state.turn == 1 and piece == 17 and j == 7:
-                is_stuck = True
+    # 先手歩
+    if np.any(np.where(bs == 1)[0] % 9 == 0):
+        is_stuck = True
+    # 先手香車
+    if np.any(np.where(bs == 2)[0] % 9 == 0):
+        is_stuck = True
+    # 先手桂馬
+    if np.any(np.where(bs == 3)[0] % 9 <= 1):
+        is_stuck = True
+    # 後手歩
+    if np.any(np.where(bs == 15)[0] % 9 == 8):
+        is_stuck = True
+    # 後手香車
+    if np.any(np.where(bs == 16)[0] % 9 == 8):
+        is_stuck = True
+    # 後手桂馬
+    if np.any(np.where(bs == 17)[0] % 9 >= 7):
+        is_stuck = True
     return is_stuck
 
 
-# ピンされた（動かすと相手の駒の利きが玉に通ってしまうので動かせない）駒の位置とピンの方向（縦横右上がり右下がり）を記録
-# ピンされた駒はピンされている方向以外には動かせない
-def _up_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
+def _direction_to_pin(direction: int) -> int:
+    if direction == 0 or direction == 5:
+        return 1
+    if direction == 1 or direction == 7:
+        return 2
+    if direction == 2 or direction == 6:
+        return 3
+    if direction == 3 or direction == 4:
+        return 4
+    return 0
+
+
+def _direction_pin(
+    bs: np.ndarray,
+    turn: int,
+    king_point: int,
+    direction: int,
+    array: np.ndarray,
 ) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    # 上方向のピン
-    u = king_point
-    u_num = 0
-    u_piece = -1
-    u_flag = False
-    # 自分より上は最大八マスしかない
-    for i in range(8):
-        # 探索位置の更新
-        u -= 1
-        # 探索が終わっている場合はスルー
-        if u_flag:
-            continue
-        # 盤外や同じ列にない場合は弾く
-        if not _is_in_board(u) or not _is_same_column(king_point, u):
-            continue
-        piece = bs[u]
-        # 駒がない場合は処理をしない
-        if piece == 0:
-            continue
-        # 駒がある場合
-        u_num += 1
-        # 1枚目
-        if u_num == 1:
-            # 自分の駒なら位置を記録
-            if _owner(piece) == turn:
-                u_piece = u
-            # 相手の駒なら関係ないので探索終了
-            else:
-                u_flag = True
-        # 2枚目
-        if u_num == 2:
-            # 先手の場合、香車、飛車、龍によってピン
-            if turn == 0:
-                if piece == 16 or piece == 20 or piece == 28:
-                    new_array[u_piece] = 1
-            # 後手の場合、飛車、龍によってピン
-            else:
-                if piece == 6 or piece == 14:
-                    new_array[u_piece] = 1
-            # 一回通ったら探索する必要はない
-            u_flag = True
-    return new_array
-
-
-def _up_left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    ul = king_point
-    ul_num = 0
-    ul_flag = False
-    for i in range(8):
-        if ul_flag:
-            continue
-        ul += 8
-        if not _is_in_board(ul) or not _is_same_declining(king_point, ul):
-            continue
-        piece = bs[ul]
-        if piece == 0:
-            continue
-        ul_num += 1
-        if ul_num == 1:
-            if _owner(piece) == turn:
-                ul_piece = ul
-            else:
-                ul_flag = True
-        if ul_num == 2:
-            if _owner(piece) != turn and (piece % 14 == 5 or piece % 14 == 13):
-                new_array[ul_piece] = 2
-            ul_flag = True
-    return new_array
-
-
-def _up_right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    ur = king_point
-    ur_num = 0
-    ur_piece = -1
-    ur_flag = False
-    for i in range(8):
-        if ur_flag:
-            continue
-        ur -= 10
-        if not _is_in_board(ur) or not _is_same_rising(king_point, ur):
-            continue
-        piece = bs[ur]
-        if piece == 0:
-            continue
-        ur_num += 1
-        if ur_num == 1:
-            if _owner(piece) == turn:
-                ur_piece = ur
-            # 相手の駒なら関係ないので探索終了
-            else:
-                ur_flag = True
-        # 2枚目
-        if ur_num == 2:
-            # 相手の角・馬によってピン
-            if _owner(piece) != turn and (piece % 14 == 5 or piece % 14 == 13):
-                new_array[ur_piece] = 3
-            ur_flag = True
-    return new_array
-
-
-def _left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    l_ = king_point
-    l_num = 0
-    l_flag = False
-    for i in range(8):
-        if l_flag:
-            continue
-        l_ += 9
-        if not _is_in_board(l_) or not _is_same_row(king_point, l_):
-            continue
-        piece = bs[l_]
-        if piece == 0:
-            continue
-        l_num += 1
-        if l_num == 1:
-            if _owner(piece) == turn:
-                l_piece = l_
-            else:
-                l_flag = True
-        # 2枚目
-        if l_num == 2:
-            if _owner(piece) != turn and (piece % 14 == 6 or piece % 14 == 0):
-                new_array[l_piece] = 4
-            l_flag = True
-    return new_array
-
-
-def _right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    r = king_point
-    r_num = 0
-    r_flag = False
-    for i in range(8):
-        if r_flag:
-            continue
-        r -= 9
-        if not _is_in_board(r) or not _is_same_row(king_point, r):
-            continue
-        piece = bs[r]
-        if piece == 0:
-            continue
-        r_num += 1
-        if r_num == 1:
-            if _owner(piece) == turn:
-                r_piece = r
-            else:
-                r_flag = True
-        # 2枚目
-        if r_num == 2:
-            if _owner(piece) != turn and (piece % 14 == 6 or piece % 14 == 0):
-                new_array[r_piece] = 4
-            r_flag = True
-    return new_array
-
-
-def _down_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    d = king_point
-    d_num = 0
-    d_piece = -1
-    d_flag = False
-    for i in range(8):
-        d += 1
-        if d_flag:
-            continue
-        if not _is_in_board(d) or not _is_same_column(king_point, d):
-            continue
-        piece = bs[d]
-        if piece == 0:
-            continue
-        # 駒がある場合
-        d_num += 1
-        # 1枚目
-        if d_num == 1:
-            # 自分の駒なら位置を記録
-            if _owner(piece) == turn:
-                d_piece = d
-            # 相手の駒なら関係ないので探索終了
-            else:
-                d_flag = True
-        # 2枚目
-        if d_num == 2:
-            # 先手の場合、香車、飛車、龍によってピン
-            if turn == 0:
-                if piece == 20 or piece == 28:
-                    new_array[d_piece] = 1
-            # 後手の場合、飛車、龍によってピン
-            else:
-                if piece == 2 or piece == 6 or piece == 14:
-                    new_array[d_piece] = 1
-            # 一回通ったら探索する必要はない
-            d_flag = True
-    return new_array
-
-
-def _down_left_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    dl = king_point
-    dl_num = 0
-    dl_flag = False
-    for i in range(8):
-        if dl_flag:
-            continue
-        dl += 10
-        if not _is_in_board(dl) or not _is_same_rising(king_point, dl):
-            continue
-        piece = bs[dl]
-        if piece == 0:
-            continue
-        dl_num += 1
-        if dl_num == 1:
-            if _owner(piece) == turn:
-                dl_piece = dl
-            else:
-                dl_flag = True
-        if dl_num == 2:
-            if _owner(piece) != turn and (piece % 14 == 5 or piece % 14 == 13):
-                new_array[dl_piece] = 3
-            dl_flag = True
-    return new_array
-
-
-def _down_right_pin(
-    bs: np.ndarray, turn: int, king_point: int, array: np.ndarray
-) -> np.ndarray:
-    new_array = copy.deepcopy(array)
-    dr = king_point
-    dr_num = 0
-    dr_piece = -1
-    dr_flag = False
-    for i in range(8):
-        if dr_flag:
-            continue
-        dr -= 8
-        if not _is_in_board(dr) or not _is_same_declining(king_point, dr):
-            continue
-        piece = bs[dr]
-        if piece == 0:
-            continue
-        dr_num += 1
-        if dr_num == 1:
-            if _owner(piece) == turn:
-                dr_piece = dr
-            # 相手の駒なら関係ないので探索終了
-            else:
-                dr_flag = True
-        # 2枚目
-        if dr_num == 2:
-            # 相手の角・馬によってピン
-            if _owner(piece) != turn and (piece % 14 == 5 or piece % 14 == 13):
-                new_array[dr_piece] = 3
-            dr_flag = True
+    new_array = array
+    dir_array = _dis_direction_array(king_point, 0, direction)
+    e_turn = (turn + 1) % 2
+    bs_one = np.where(bs == 0, 0, 1)
+    dir_one_array = dir_array * bs_one
+    if np.count_nonzero(dir_one_array) <= 1:
+        return new_array
+    dif = _direction_to_dif(direction, 0)
+    dis1 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 0)[0]
+    dis2 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 1)[1]
+    point1 = king_point + dis1 * dif
+    point2 = king_point + dis2 * dif
+    piece1 = bs[point1]
+    piece2 = bs[point2]
+    if _owner(piece1) != turn:
+        return new_array
+    flag = False
+    if direction == 0:
+        if turn == 0 and piece2 == 16:
+            flag = True
+    if direction == 5:
+        if turn == 1 and piece2 == 2:
+            flag = True
+    if direction == 0 or direction == 3 or direction == 4 or direction == 5:
+        if piece2 == 6 + e_turn * 14 or piece2 == 14 + e_turn * 14:
+            flag = True
+    if direction == 1 or direction == 2 or direction == 6 or direction == 7:
+        if piece2 == 5 + e_turn * 14 or piece2 == 13 + e_turn * 14:
+            flag = True
+    if flag:
+        new_array[point1] = _direction_to_pin(direction)
     return new_array
 
 
@@ -1306,20 +1127,14 @@ def _pin(state: ShogiState):
     turn = state.turn
     pins = np.zeros(81, dtype=np.int32)
     king_point = int(state.board[8 + 14 * turn, :].argmax())
-    pins = _up_pin(bs, turn, king_point, pins)
-    pins = _up_left_pin(bs, turn, king_point, pins)
-    pins = _up_right_pin(bs, turn, king_point, pins)
-    pins = _left_pin(bs, turn, king_point, pins)
-    pins = _right_pin(bs, turn, king_point, pins)
-    pins = _down_pin(bs, turn, king_point, pins)
-    pins = _down_left_pin(bs, turn, king_point, pins)
-    pins = _down_right_pin(bs, turn, king_point, pins)
+    for i in range(8):
+        pins = _direction_pin(bs, turn, king_point, i, pins)
     return pins
 
 
 # 特定の方向の動きだけを除いたactionを生成する
 def _eliminate_direction(actions: np.ndarray, direction: int):
-    new_array = copy.deepcopy(actions)
+    new_array = actions
     # 2方向と成・不成の4方向のフラグを折る
     dir1 = 0
     dir2 = 0
@@ -1341,11 +1156,10 @@ def _eliminate_direction(actions: np.ndarray, direction: int):
         dir2 = 4
     pro_dir1 = dir1 + 8
     pro_dir2 = dir2 + 8
-    for i in range(81):
-        new_array[dir1 * 81 + i] = 0
-        new_array[dir2 * 81 + i] = 0
-        new_array[pro_dir1 * 81 + i] = 0
-        new_array[pro_dir2 * 81 + i] = 0
+    new_array[dir1 * 81 : (dir1 + 1) * 81] = 0
+    new_array[dir2 * 81 : (dir2 + 1) * 81] = 0
+    new_array[pro_dir1 * 81 : (pro_dir1 + 1) * 81] = 0
+    new_array[pro_dir2 * 81 : (pro_dir2 + 1) * 81] = 0
     return new_array
 
 
@@ -1373,7 +1187,7 @@ def _king_escape(state: ShogiState) -> bool:
     effects = _kingless_effected_positions(
         bs, king_point, _another_color(state)
     )
-    king_moves = _small_piece_moves(8, king_point)
+    king_moves = POINT_MOVES[king_point][8]
     flag = False
     for i in range(81):
         if (
@@ -1386,19 +1200,12 @@ def _king_escape(state: ShogiState) -> bool:
 
 
 def _between(point1: int, point2: int) -> np.ndarray:
-    flag = True
     between = np.zeros(81, dtype=np.int32)
     direction = _point_to_direction(point1, point2, False, 0)
     if direction == -1:
         return between
     dif = _direction_to_dif(direction, 0)
-    p = point1
-    for i in range(7):
-        p += dif
-        if p == point2:
-            flag = False
-        if flag:
-            between[p] = 1
+    np.put(between, np.arange(point1 + dif, point2, dif), 1)
     return between
 
 
